############################################################################
#                                                                          #
# Imports & Constants                                                      #
#                                                                          #
############################################################################

import gpudb
import csv
from tabulate import tabulate
from collections import OrderedDict

HOST = "127.0.0.1"
PORT = "9191"
CSV = "road_weights.csv"

OPTION_NO_ERROR = {"no_error_if_not_exists": "true"}

TABLE_SRN = "seattle_road_network"

GRAPH_S = TABLE_SRN + "_graph"
GRAPH_S_SPSOLVED = GRAPH_S + "_shortest_path_solved"
GRAPH_S_SPSOLVED2 = GRAPH_S + "_shortest_path_solved2"
GRAPH_S_SPSOLVED3 = GRAPH_S + "_shortest_path_solved3"


############################################################################
#                                                                          #
# Table Setup Function                                                     #
#                                                                          #
############################################################################

def table_setup():
    """ Setup necessary source tables for graph solver examples. """

    print("")
    print("===========")
    print("TABLE SETUP")
    print("===========")
    print("")

    # Clear any related tables in case they already exist
    kinetica.clear_table(table_name=TABLE_SRN, options=OPTION_NO_ERROR)

    # Define the Seattle road network table type
    srn_columns = [
        ["OriginalEdgeID", "long"],
        ["TwoWay", "int"],
        ["WKTLINE", "string", "wkt"],
        ["time", "float"]
    ]

    # Create the Seattle road network table
    try:
        table_srn_obj = gpudb.GPUdbTable(
            _type=srn_columns,
            name=TABLE_SRN,
            db=kinetica,
            options={}
        )
        print("{} table object successfully created.".format(TABLE_SRN))
    except gpudb.GPUdbException as e:
        print("{} table object creation failure: {}".format(TABLE_SRN, str(e)))

    # Insert records from a local CSV file into the Seattle road network table
    print("Creating records from the " + CSV + " file.")
    srn_data = csv.reader(open(CSV))
    srn_data.next()
    srn_records = []
    for record in srn_data:
        record_data = [int(record[0]), int(record[1]), record[2],
                       float(record[3])]
        srn_records.append(record_data)
    print(
        "Inserting records into {} table. This may take a "
        "while.".format(TABLE_SRN)
    )
    table_srn_obj.insert_records(srn_records)
    print(
        "Number of records inserted into the {} "
        "table: {}".format(TABLE_SRN, table_srn_obj.size())
    )

# end table_setup()

############################################################################
#                                                                          #
# Graph Setup Function                                                     #
#                                                                          #
############################################################################

def graph_setup():
    """ Setup graphs to be used in solver examples. """

    print("")
    print("===========")
    print("GRAPH SETUP")
    print("===========")
    print("")

    # Create a graph from TABLE_SRN
    print("Creating {}".format(GRAPH_S))
    create_s_graph_response = kinetica.create_graph(
        graph_name=GRAPH_S,
        directed_graph=True,
        nodes=[],
        edges=[
            TABLE_SRN + ".WKTLINE AS EDGE_WKTLINE",
            TABLE_SRN + ".TwoWay AS EDGE_DIRECTION"
        ],
        weights=[
            TABLE_SRN + ".WKTLINE AS WEIGHTS_EDGE_WKTLINE",
            TABLE_SRN + ".TwoWay AS WEIGHTS_EDGE_DIRECTION",
            TABLE_SRN + ".time AS WEIGHTS_VALUESPECIFIED"
        ],
        restrictions=[],
        options={
            "recreate": "true",
            "enable_graph_draw": "true",
            "graph_table": "seattle_graph_debug"
        }
    )
    print(
        "{} created: {}".format(GRAPH_S,
                                create_s_graph_response["status_info"]
                                ["status"])
    )
    print("Number of nodes: {}".format(create_s_graph_response["num_nodes"]))
    print("Number of edges: {}".format(create_s_graph_response["num_edges"]))

# end graph_setup()

############################################################################
#                                                                          #
# Shortest Path Solving Function                                           #
#                                                                          #
############################################################################

def shortest_path_example():
    """Demonstrate creating and solving a graph using the SHORTEST_PATH method.
    """

    print("")
    print("==============================")
    print("SHORTEST PATH SOLVER EXAMPLE 1")
    print("==============================")
    print("")

    # Clear any related tables in case they already exist
    kinetica.clear_table(
        table_name=GRAPH_S_SPSOLVED,
        options=OPTION_NO_ERROR
    )

    # Solve GRAPH_S for Shortest Path using a single given source node and 
    # routing to a single given destination node
    source_nodes = ["POINT(-122.1792501 47.2113606)"]
    destination_nodes = ["POINT(-122.2221 47.5707)"]
    solve_s_spgraph_response = kinetica.solve_graph(
        graph_name=GRAPH_S,
        solver_type="SHORTEST_PATH",
        source_nodes=source_nodes,
        destination_nodes=destination_nodes,
        solution_table=GRAPH_S_SPSOLVED,
        options={"export_solve_results": "true"}
    )["result_per_destination_node"]
    print("Cost per destination node when source node = {}: ".format(
        source_nodes
    ))

    # Divide the individual destination costs by 60 to convert to minutes
    solve_s_spgraph_response[:] = [
        cost / 60 for cost in solve_s_spgraph_response
    ]

    # Display the results
    headers = ["Destination Node", "Cost (in minutes)"]
    print(
        tabulate(
            {
                headers[0]: destination_nodes,
                headers[1]: solve_s_spgraph_response
            },
            headers="keys",
            tablefmt="grid"
        )
    )

    print("")
    print("==============================")
    print("SHORTEST PATH SOLVER EXAMPLE 2")
    print("==============================")
    print("")

    # Clear any related tables in case they already exist
    kinetica.clear_table(
        table_name=GRAPH_S_SPSOLVED2,
        options=OPTION_NO_ERROR
    )

    # Solve GRAPH_S for Shortest Path using the given source node and routing to
    # the given destination nodes
    source_nodes = ["POINT(-122.1792501 47.2113606)"]
    destination_nodes = [
        "POINT(-122.2221 47.5707)", 
        "POINT(-122.541017 47.809121)",
        "POINT(-122.520440 47.624725)", 
        "POINT(-122.467915 47.427280)"
    ]
    solve_s_spgraph_response2 = kinetica.solve_graph(
        graph_name=GRAPH_S,
        solver_type="SHORTEST_PATH",
        source_nodes=source_nodes,
        destination_nodes=destination_nodes,
        solution_table=GRAPH_S_SPSOLVED2,
        options={"export_solve_results": "true"}
    )["result_per_destination_node"]
    print("Cost per destination node when source node = {}: ".format(
        source_nodes
    ))

    # Divide the individual destination costs by 60 to convert to minutes
    solve_s_spgraph_response2[:] = [
        cost / 60 for cost in solve_s_spgraph_response2
    ]

    # Display the results
    headers = ["Destination Node", "Cost (in minutes)"]
    print(
        tabulate(
            {
                headers[0]:destination_nodes, 
                headers[1]:solve_s_spgraph_response2
            }, 
            headers="keys", 
            tablefmt="grid"
        )
    )

    print("")
    print("==============================")
    print("SHORTEST PATH SOLVER EXAMPLE 3")
    print("==============================")
    print("")

    # Clear any related tables in case they already exist
    kinetica.clear_table(
        table_name=GRAPH_S_SPSOLVED3,
        options=OPTION_NO_ERROR
    )

    # Solve GRAPH_S for Shortest Path using the given source nodes and routing 
    # to the given destination nodes
    source_nodes = [
        "POINT(-122.1792501 47.2113606)",
        "POINT(-122.1792501 47.2113606)",
        "POINT(-122.375180125237 47.8122103214264)",
        "POINT(-122.375180125237 47.8122103214264)"
    ]
    destination_nodes = [
        "POINT(-122.2221 47.5707)",
        "POINT(-122.541017 47.809121)",
        "POINT(-122.520440 47.624725)",
        "POINT(-122.467915 47.427280)"
    ]
    solve_s_spgraph_response3 = kinetica.solve_graph(
        graph_name=GRAPH_S,
        solver_type="SHORTEST_PATH",
        source_nodes=source_nodes,
        destination_nodes=destination_nodes,
        solution_table=GRAPH_S_SPSOLVED3,
        options={"export_solve_results": "true"}
    )["result_per_destination_node"]
    print("Cost per destination node: ")

    # Divide the individual destination costs by 60 to convert to minutes
    solve_s_spgraph_response3[:] = [
        cost / 60 for cost in solve_s_spgraph_response3
    ]

    # Display the results
    headers = ["Source Node", "Destination Node", "Cost (in minutes)"]
    rows = []
    for num_nodes in range(0, len(destination_nodes)):
        rows.append(
            OrderedDict(
                [
                    (headers[0], source_nodes[num_nodes]), 
                    (headers[1], destination_nodes[num_nodes]), 
                    (headers[2], solve_s_spgraph_response3[num_nodes])
                ]
            )
        )
    print(
        tabulate(
            rows,
            headers="keys",
            tablefmt="grid"
        )
    )

# end shortest_path_example()

if __name__ == '__main__':
    # Establish connection with a locally-running instance of Kinetica
    kinetica = gpudb.GPUdb(encoding='BINARY', host=HOST, port=PORT)

    # Execute defined functions
    table_setup()
    graph_setup()
    shortest_path_example()
